/***********************************************************************
 **  class Point: 2D point
 ***********************************************************************/

#if !defined(POINT_H)
#define POINT_H

#include <vector>

namespace stdi { 

  class  Point  {
  
  public:

    // Typedefs
    typedef short coord_type;
    
    Point(): xp(0), yp(0) {}
    Point(int xpos, int ypos):xp((coord_type)xpos), yp((coord_type)ypos) {}
    ~Point(){}
    
    bool isNull() const { return xp == 0 && yp == 0; }
    int x() const { return xp; }
    int y() const { return yp; }

    void setX( int x ) { xp = (coord_type)x; }
    void setY( int y ) { yp = (coord_type)y; }
  
    Point &operator+=( const Point &p );
    Point &operator-=( const Point &p );

    Point &operator+=( int c );
    Point &operator+=( double c );

    Point &operator-=( int c );
    Point &operator-=( double c );

    Point &operator*=( int c );
    Point &operator*=( double c );

    Point &operator/=( int c );
    Point &operator/=( double c );
    
    friend bool operator==( const Point &, const Point & );
    friend bool operator<( const Point &p1, const Point &p2 );
    friend bool operator!=( const Point &, const Point & );
    friend const Point operator+( const Point &, const Point & );
    friend const Point operator-( const Point &, const Point & );
    friend const Point operator-( const Point & );

    // point + scalar, scalar + point
    friend const Point operator+( const Point &p, int c );
    friend const Point operator+( int c, const Point &p );
    friend const Point operator+( const Point &p, double c );
    friend const Point operator+( double c, const Point &p );

    // point - scalar, scalar - point
    friend const Point operator-( const Point &p, int c );
    friend const Point operator-( int c, const Point &p );
    friend const Point operator-( const Point &p, double c );
    friend const Point operator-( double c, const Point &p );


    // point * scalar, scalar * point
    friend const Point operator*( const Point &, int );
    friend const Point operator*( int, const Point & );
    friend const Point operator*( const Point &, double );
    friend const Point operator*( double, const Point & );

    // point / scalar
    friend const Point operator/( const Point &, int );
    friend const Point operator/( const Point &, double );
  
  private:

    coord_type xp, yp;
  };

  typedef  std::vector<Point> PointArray;

} // End of namespace stdi

#endif // POINT_H



