#include "bip.h"
#include "global.h"
#include "point.h"
#include "point3d.h"
#include "image_io.h"
#include "scalespace.h"

using namespace stdi;
using namespace bip;

#include <iostream>
#include <string>
#include <cmath>
#include <algorithm>
using namespace std;
#define MAXVAL 10000

typedef IntArray::iterator IntArrayIterator;

namespace { // Begin unnamed namespace

  // Local functions
  template<class T>
  bool isEdge(const image<T> &img, int x, int y, int z) {
    if (img.isNull()) return false;
    int x1 = min(img.nx()-1, x+1);
    int y1 = min(img.ny()-1, y+1);
    int z1 = min(img.nz()-1, z+1);
    int x_1 = max(0, x-1); 
    int y_1 = max(0, y-1);
    int z_1 = max(0, z-1);  
    int xx, yy, zz;
    T v = img.at(x, y, z);
    for(zz = z_1; zz <= z1; zz++)
      for(yy = y_1; yy <= y1; yy++)
	for(xx = x_1; xx <= x1; xx++)
	  if (img.at(xx, yy, zz) != v) 
	    return true;
    return false;
  }
  
  template<class T>
  bool isEdge(const image<T> &img, int x, int y) {
    if (img.isNull()) return false;
    int x1 = min(img.nx()-1, x+1);
    int y1 = min(img.ny()-1, y+1);
    int x_1 = max(0, x-1); 
    int y_1 = max(0, y-1);    
    int xx, yy;
    T v = img.at(x, y);
  
    for(yy = y_1; yy <= y1; yy++)
      for(xx = x_1; xx <= x1; xx++)
	if (img.at(xx, yy) != v) 
	  return true;
    return false;
  }

  template<class T>
  ByteImage __edge_2d__(const image<T> &img) {
    if (img.isNull()) return ByteImage();
    const int nx ( img.nx() );
    const int ny ( img.ny() );
    const int nx_1 (nx-1);
    const int ny_1 (ny-1);
    int n, x, y, xx, yy;
    ByteImage b(nx, ny);
    b.reset(Background);

    T br;

    for(y=1; y<ny_1; y++)
      for(x=1; x<nx_1; x++)  {
	br = img.at(x, y);
	n = 0;
	for(yy=y-1; yy<=y+1; yy++)
	  for(xx=x-1;xx<=x+1;xx++)
	    if (img.at(xx,yy)==br) n++;
	if(n<8) b.set(x, y, Foreground);
      }
    return b;
  }


  FloatImage __init__chamfer_byte__(const ByteImage &img) {
  
    //  cout << "Byte chamfer" << endl;
    const int npix ( img.npix() );
    FloatImage D(img.nx(), img.ny(), img.nz());
    for (int i=0; i<npix; i++) 
      D.set(i, img.at(i)==Background? 0 : MAXVAL );
    return D;
  }


  FloatImage __init__chamfer_short__(const ShortImage &img) 
  {  
    //  const int npix( img.npix() );
    const int nx( img.nx() );
    const int ny( img.ny() );
    const int nz( img.nz() );
    FloatImage D(nx, ny, nz);
    int x, y, z;
  
    if (img.nz()>1) {
      for(z=0; z<nz; z++)
	for(y=0; y<ny; y++)
	  for(x=0; x<nx; x++) 
	    if (img.at(x, y, z)<0) 
	      D.set(x, y, z, 0);	
	    else if (isEdge(img, x, y, z)) 
	      D.set(x, y, z, 1);
	    else D.set(x, y, z, MAXVAL);
    } else 
      for(y=0; y<ny; y++)
	for(x=0; x<nx; x++) 
	  if (img.at(x, y)<0) 
	    D.set(x, y, 0);	
	  else if (isEdge(img, x, y)) 
	    D.set(x, y, 1);
	  else D.set(x, y, MAXVAL);
    return D;
  }

  void __compute_chamfer_3d__(FloatImage &D) {

    if (D.isNull()) return;

    const int nc = D.nx(); 
    const int nr = D.ny(); 
    const int ni = D.nz(); 
    const int nc_1 = nc-1;
    const int nr_1 = nr-1; 
    const int ni_1 = ni-1;
    const int nc_2 = nc-2;
    const int nr_2 = nr-2;
    const int ni_2 = ni-2;
    const int npix = D.npix();
    const float max_val(MAXVAL);
    float min_val, d;
    int  i, j, z, k, p;


    IntArray off;
    vector<float> off_dist;
    //  IntArrayIterator it;

    off = IntArray();
    off_dist = vector<float>();
    off_dist.reserve(27);
    off.reserve(27);

    for(z = -1; z <= 1; z++)
      for(i = -1; i <= 1; i++)
	for(j = -1; j <= 1; j++) 
	  if ((k = z*nr*nc + i*nc+j) != 0 && i+j+z<=0 ) {
	    off.push_back(k);
	    d = sqrt( float( z*z+i*i+j*j) );
	    off_dist.push_back(d);
	  }

    for(z = 1; z < ni_1; z++) 
      for (i = 1; i < nr_1; i++)
	for (j = 1; j < nc_1; j++) 
	  if (  D.pixel(p = D.pos(j, i, z)) > 0.5 ){	  
	    min_val = max_val + 2;
	    for(k = 0; k < off.size(); k++)
	      if ( (d = D.pixel(p + off[k])+off_dist[k]) < min_val)
		min_val = d;
	  
	    if (min_val >= max_val) continue; 
	    if (min_val  < D.pixel(p))
	      D.set(p, min_val);
	  }
  
    off = IntArray();
    off.reserve(27);
    off_dist = vector<float>();
    off_dist.reserve(27);

    for(z = -1; z <= 1; z++)
      for(i = -1; i <= 1; i++)
	for(j = -1; j <= 1; j++) 
	  if ((k = z*nr*nc + i*nc+j) != 0 && i+j+z>=0) {
	    off.push_back(k);
	    d = sqrt(float(z*z+i*i+j*j));
	    off_dist.push_back(d);
	  }
  

    for(z = ni_2; z > 0; z--)
      for (i = nr_2; i > 0; i--)
	for (j = nc_2; j > 0; j--) 
	  if (  D.pixel(p=D.pos(j, i, z)) > 0.5 ){	  
	    min_val = max_val + 2;
	    for(k = 0; k < off.size(); k++)
	      if ( (d = D.pixel(p + off[k])+off_dist[k]) < min_val)
		min_val = d;
	  
	    if (min_val >= max_val) continue; 
	    if (min_val  < D.pixel(j, i, z))
	      D.set(p, min_val);
	  }

    for(z=0; z<ni; z++)
      for(i=0; i<nr; i++)
	for(j=0; j<nc; j++)
	  if (D(j, i, z)>=max_val-1)
	    for(int zz=z-1; zz<=z+1; zz++)
	      for(int ii=i-1; ii<=i+1; ii++)
		for(int jj=j-1; jj<=j+1; jj++)
		  if (D.inImage(jj, ii, zz))
		    if (D(jj, ii, zz)+1<D(j, i, z))
		      D(j, i, z) = D(jj, ii, zz) + 1;
  
  }



  void __compute_chamfer_2d__(FloatImage &D) 
  {
    if (D.isNull()) return;
    const float max_val(MAXVAL);
    const int nc = D.nx(); 
    const int nr = D.ny(); 
    const int nc_1 = nc-1;
    const int nr_1 = nr-1; 
    const int nc_2 = nc-2;
    const int nr_2 = nr-2;
    const int npix = D.npix();
    int  i, j,  k, p;
    float min_val;
  
    IntArray off;
    vector<float> off_dist;
    float d;
  
    // cout << "1st scan..." ; 
    off = IntArray();
    off_dist = vector<float>();
    off_dist.reserve(8);
    off.reserve(8);

    for(i = -1; i <= 1; i++)
      for(j = -1; j <= 1; j++) 
	if ((k = i*nc+j) != 0 && i+j<=0) {
	  off.push_back(k);
	  d = sqrt(float(i*i+j*j));
	  off_dist.push_back(d);
	}
  
    for (i = 1; i < nr_1; i++)
      for (j = 1; j < nc_1; j++) 
	if (  D.pixel(p = D.pos(j, i)) > 1 ){	  
	  min_val = max_val + 2;
	  for(k = 0; k < off.size(); k++)
	    if ( (d = D.pixel(p + off[k])+off_dist[k]) < min_val)
	      min_val = d;
	
	  if (min_val >= max_val) continue; 
	  if (min_val  < D.pixel(p))
	    D.set(p, min_val);
	}
  
    // cout << "2nd scan..." ;
    off = IntArray();
    off.reserve(8);
    off_dist = vector<float>();
    off_dist.reserve(8);


    for(i = -1; i <= 1; i++)
      for(j = -1; j <= 1; j++) 
	if ((k = i*nc+j) != 0 && i+j>=0 ) {
	  off.push_back(k);
	  d = sqrt(float(i*i+j*j));
	  off_dist.push_back(d);
	}
  
    for (i = nr_2; i > 0; i--)
      for (j = nc_2; j > 0; j--) 
	if (  D.pixel(p=D.pos(j, i)) > 1 ){	  
	  min_val = max_val + 2;
	  for(k = 0; k < off.size(); k++)
	    if ( (d = D.pixel(p + off[k])+off_dist[k]) < min_val)
	      min_val = d;
	
	  if (min_val >= max_val) continue; 
	  if (min_val  < D.pixel(j, i))
	    D.set(p, min_val);
	}
  
    for(i=0; i<nr; i++)
      for(j=0; j<nc; j++)
	if (D(j, i)>=max_val-1)
	  for(int ii=i-1; ii<=i+1; ii++)
	    for(int jj=j-1; jj<=j+1; jj++)
	      if (D.inImage(jj, ii))
		if (D(jj, ii)+1<D(j, i))
		  D(j, i) = D(jj, ii) + 1;
  }



  void __remove__internal_structures__(ByteImage &img)
  {
    int nL;
    ShortImage L = bip::regions(img, nL, Background);
    vector<RegionFeature> rf = regionFeatures(L, nL);
    int lmax = -1;
    int vmax = -1;
    int v;

    for(int i=0; i<rf.size(); i++) 
      if ((v=rf[i].volume())>vmax) {
	lmax = i;
	vmax = v; 
      }
    if (lmax < 0) return;
    const int npix = img.npix();
    for(int i=0; i<npix; i++)
      img.set(i, L.at(i)==lmax? Background : Foreground);   
  }


} // End of unnamed namespace


bool bip::isBinaryImage(const Image &I)
{
  if (I.format()!=BYTE) return false;
  ByteImage *img = (ByteImage *) &I;
  const int npix = img->npix();
  for(int i=0; i<npix; i++)
    if (img->at(i) != Foreground && img->at(i) != Background )
      return false;
  return true;
}


// RTTI is used here, the reasons I do not use virtual function:
// (1) bip:: edge is not used extensively, and (2) to keep the Image
// class as small as possible
ByteImage bip::edge(const Image &img)
{
  try {
    switch(img.format()) {
    case BYTE:
      return __edge_2d__(dynamic_cast<const ByteImage&>(img));
    case SHORT:
      return __edge_2d__(dynamic_cast<const ShortImage&>(img));
    case INT:
      return __edge_2d__(dynamic_cast<const IntImage&>(img));
    case FLOAT:
      return __edge_2d__(dynamic_cast<const FloatImage&>(img));
    default:
      return ByteImage();
    }
  } catch (bad_cast) {
    return ByteImage();
  }
}


ShortImage bip::regions(const ByteImage &img)
{
  int nc;
  return bip::regions(img, nc,	byte(Foreground));
}

ShortImage bip::regions(const ByteImage &img, int &nlabel,
			int intensity, int radius, stdi::NeighborhoodType nb )
{  
  if (img.isNull()) return ShortImage();
 
  const int nx = img.nx();
  const int ny = img.ny();
  const int nz = img.nz();
  const int npix = img.npix();
  const int nx_1 = nx-1;
  const int ny_1 = ny-1;
  const int nz_1 = nz-1;
  
  ShortImage label(nx, ny, nz);
  if (intensity < 0) 
    label.reset(-2);
  else 
    for(int i = 0; i < npix; i++)
      label.set(i, (img.at(i) != intensity) ? -1 : -2);
  
  int x, y, z,  k, x_1, x0, x1, y_1, y0, y1, z_1, z0, z1, xx, yy, zz; 
  int cur_int;
  queue <Point3D> Q;
  Point3D q;
  
  nlabel = 0;
  for(z = 0; z < nz; z++)     
    for(y = 0; y < ny; y++) 
      for(x = 0; x < nx; x++) 
	if (label.at(x, y, z) == -2) {
	  
	  Q = queue<Point3D>();
	  Q.push(Point3D(x,y,z));
	  cur_int = img.at(x, y, z); 
	  while (!Q.empty()) {
	    
	    q = Q.front();
	    xx = q.x(); yy = q.y(); zz = q.z();
	    label.set(xx, yy, zz, nlabel);
	    
	    if ( (x_1 = xx - radius) < 0 ) x_1 = 0;
	    if ( (x1 = xx + radius) >= nx ) x1 = nx_1;
	    if ( (y_1 = yy - radius) < 0 ) y_1 = 0;
	    if ( (y1 = yy + radius) >= ny ) y1 = ny_1;
	    if ( (z_1 = zz - radius) < 0 ) z_1 = 0;
	    if ( (z1 = zz + radius) >= nz ) z1 = nz_1;	    
	    
	    if( nb == stdi::Eight ){
		for(z0 = z_1; z0 <= z1; z0++)
		    for(y0 = y_1; y0 <= y1; y0++)
			for(x0 = x_1; x0 <= x1; x0++)
			    if (label.at(x0, y0, z0) == -2 && img.at(x0, y0, z0)==cur_int) {   
				Q.push(Point3D(x0, y0, z0));
				label.set(x0, y0, z0, nlabel);
			    }	      	    
	    }
	    else if( nb == stdi::Four ){
		
		if( label.at(xx,y_1,zz) == -2 && img.at(xx,y_1,zz) == cur_int ){
		    Q.push(Point3D(xx,y_1,zz));
		    label.set(xx,y_1,zz,nlabel);
		}
		if( label.at(xx,y1,zz) == -2 && img.at(xx,y1,zz) == cur_int ){
		    Q.push(Point3D(xx,y1,zz));
		    label.set(xx,y1,zz,nlabel);
		}
		if( label.at(x_1,yy,zz) == -2 && img.at(x_1,yy,zz) == cur_int ){
		    Q.push(Point3D(x_1,yy,zz));
		    label.set(x_1,yy,zz,nlabel);
		}
		if( label.at(x1,yy,zz) == -2 && img.at(x1,yy,zz) == cur_int ){
		    Q.push(Point3D(x1,yy,zz));
		    label.set(x1,yy,zz,nlabel);
		}
		if( label.at(xx,yy,z_1) == -2 && img.at(xx,yy,z_1) == cur_int ){
		    Q.push(Point3D(xx,yy,z_1));
		    label.set(xx,yy,z_1,nlabel);
		}
		if( label.at(xx,yy,z1) == -2 && img.at(xx,yy,z1) == cur_int ){
		    Q.push(Point3D(xx,yy,z1));
		    label.set(xx,yy,z1,nlabel);
		}
		
	    }
	    
	    Q.pop();
	    
	  }  
	  
	  nlabel++;
	} 

  return label;
}

ByteImage bip::threshold(const Image &I, float thr) {
  if (I.isNull()) return ByteImage();
  ByteImage E(I.nx(), I.ny(), I.nz());    
  const int npix ( I.npix() );
  for (int i = 0; i < npix; i++) 
    E.set(i, I.pixel(i)>=thr ? Foreground : Background );    
  return E;
}


float bip::computeThreshold(const Image &I, float scale) 
{
  if (I.isNull()) {
    cerr << "Warning: " << endl 
	 << "In bip::computeThreshold, input image is empty!" << endl 
	 << "0 returned" << endl ;
    return 0;
  }

  ByteImage e = scalespace::features(I, scale, BrightZC);
  float smin = 0, smax = 0;
  int nmin = 0, nmax = 0;
  const int npix = e.npix();
  for(int i=0; i<npix; i++)
    if (e(i)==Foreground) {
      nmax++;
      smax += I.pixel(i);
    } else {
      nmin++;
      smin += I.pixel(i);      
    }
  
  if (nmax>0) smax /= nmax;
  if (nmin>0) smin /= nmin;
  
  return (smax+smin)/2.0;
}

int bip::regionNumber(const ShortImage &img)
{
  if (img.isNull()) return 0;
  int l = img.max();
  if (l < 0) return 0;
  return int(l+1);
}


vector<RegionFeature> bip::regionFeatures(const ShortImage &label,
					  int nlabel,
					  const Image &br)
{  
  if (label.isNull()) return  vector<RegionFeature>();

  const int nx ( label.nx() );
  const int ny ( label.ny() );
  const int nz ( label.nz() );
  const int npix ( label.npix() );
  vector<RegionFeature> rf(nlabel, RegionFeature());
 
  int x, y, z, k;  
  for(k=0; k<nlabel; k++) {
    rf[k].xc = 0; rf[k].yc = 0; rf[k].zc = 0;  rf[k].radius = 0;
    rf[k].npix = 0; rf[k].mean = 0; rf[k].var = 0;
  }
  
  for(z = 0; z < nz; z++)
    for(y = 0; y < ny; y++)
      for(x = 0; x < nx; x++) 
	if ( (k=label.at(x, y, z)) >= 0 && k < nlabel) {
	  if (rf[k].npix == 0) {
	    rf[k].xmin = rf[k].xmax = x;
	    rf[k].ymin = rf[k].ymax = y;
	    rf[k].zmin = rf[k].zmax = z;	    
	  } else {
	    rf[k].xmin = min(rf[k].xmin, x);
	    rf[k].xmax = max(rf[k].xmax, x);
	    rf[k].ymin = min(rf[k].ymin, y);
	    rf[k].ymax = max(rf[k].ymax, y);
	    rf[k].zmin = min(rf[k].zmin, z);
	    rf[k].zmax = max(rf[k].zmax, z);
	  }
	  rf[k].xc += x; rf[k].yc += y; rf[k].zc += z;
	  rf[k].npix++; 
	}
  
  float radius;
  for(k=0; k<nlabel; k++) 
    if (rf[k].npix > 0) {
      rf[k].xc /= rf[k].npix; 
      rf[k].yc /= rf[k].npix; 
      rf[k].zc /= rf[k].npix;
      if (nz==1) // 2D
	radius = sqrt((float)rf[k].npix/3.14159);
      else // 3D
	radius = std::pow( rf[k].npix*3.0/(4.0*3.14159), 1./3.);
      rf[k].radius = max((int)(radius), 1);
    }
  
  // If intensity(brightness) image is given, 
  // then compute the following statistical features 
  if (br.isNull()) return rf; // If no intensity image is given
  if (!label.equalSize(br)) {
    cerr << "In bip::regionFeatures" << endl;
    cerr << "   Warning: label and brightness must have the same size" << endl;
    cerr << "   Brightness image ignored" << endl;
    return rf;
  }

  for(k = 0; k < npix; k++)
    if (label.at(k) >= 0)  
      rf[label.at(k)].mean += br.pixel(k);    
  
  for(k = 0; k < nlabel; k++)
    if (rf[k].npix > 0)
      rf[k].mean /= (float) rf[k].npix;
  
  float d;
  for(int i=0; i<npix; i++)
    if ((k = label.at(i)) >= 0 && k < nlabel) {
      d = br.pixel(i) -  rf[k].mean;
      rf[k].var += d*d;
    }
  
  for(k = 0; k < nlabel; k++)
    if (rf[k].npix > 1)
      rf[k].var =  rf[k].var/ (float) (rf[k].npix-1);
    else rf[k].var = 0;

  return rf;
}


vector<RegionFeature> bip::regionFeatures(const ShortImage &label, 
					  int nlabel)
{
  return regionFeatures(label, nlabel, ByteImage());
}

vector<RegionFeature> bip::regionFeatures(const ShortImage &label, const Image &br)
{
  return regionFeatures(label, regionNumber(label), br);
}

vector<RegionFeature> bip::regionFeatures(const ShortImage &label)
{
  return regionFeatures(label, regionNumber(label), ByteImage());
}


void __convex_hull_2d__(ShortImage &label, int nlabel, const int N)
{
  vector<RegionFeature> rf = bip::regionFeatures(label, nlabel);
  
  int  i, j, k, n, x, y;
  int xmin, xmax, ymin, ymax;
  const float pi = 3.1415927;
  const float delta = pi/(float)N;
  
  vector<float> Bmin(N), Bmax(N), nx(N), ny(N);
  float v;
  
  for(i = 0; i<N; i++) {
    nx[i] = cos(i*delta);
    ny[i] = sin(i*delta);   
  }
  
  bool added;
  const float bignum = 999999; 
  cout << nlabel << " regions detected." << endl;
  
  for(int l = 0; l < nlabel; l++) {

    ymin = rf[l].ymin; ymax = rf[l].ymax;
    xmin = rf[l].xmin; xmax = rf[l].xmax;

    for(k=0; k<N; k++){ Bmin[k] = bignum; Bmax[k] = -bignum; }

    // First scan, compute bounds
    for (y = ymin; y <= ymax; y++)
      for (x = xmin; x <= xmax; x++)
	if (label.at(x,y)==l && isEdge(label, x, y) )
	  for(i = 0; i < N; i++) {
	    v = nx[i]*x + ny[i]*y;
	    Bmin[i] = min(v, Bmin[i]);
	    Bmax[i] = max(v, Bmax[i]);
	  }
    
    // Second scan
    for (y = ymin; y <= ymax; y++)
      for (x = xmin; x <= xmax; x++)
	if (label.at(x,y) < 0 ) {
	  for(i = 0, added = true; i < N; i++)
	    if ((v = nx[i]*x + ny[i]*y) < Bmin[i] || v > Bmax[i]) {
	      added = false;
	      break;
	    }
	  if (added) label.set(x,y,l);
	} 
  }

}


ShortImage __convex_hull_2d__(const ByteImage &img, const int N)
{
  if (img.isNull()) return ShortImage();
  int nlabel;
  ShortImage label = bip::regions(img, nlabel);
  __convex_hull_2d__(label, nlabel, N);
  return label;
}

void __convex_hull_3d__(ShortImage &label, int nlabel, const int N)
{
  vector<RegionFeature> rf = bip::regionFeatures(label, nlabel);
 
  int  i, j, k, n, x, y, z;
  int xmin, xmax, ymin, ymax, zmin, zmax;
  const float pi = 3.1415927;
  const int ntheta = 4*N;
  const int nphi = N + 1;
  const int nbounds = ntheta*nphi;
  const float delta = pi/(2.0*(float)N);
 
  vector<float> Bmin(nbounds), Bmax(nbounds),
    nx(nbounds), ny(nbounds), nz(nbounds);
  float v;
 
  for(i = 0, k = 0; i < nphi; i++)
    for(j = 0; j < ntheta; j++, k++) {
      nx[k] = cos(j*delta) * sin(i*delta);
      ny[k] = sin(j*delta) * sin(i*delta);
      nz[k] = cos(i*delta);
    }
  bool added;
  const float bignum = 999999; 
  cout << nlabel << " regions detected." << endl;

  for(int l = 0; l < nlabel; l++) {
    zmin = rf[l].zmin; zmax = rf[l].zmax; 
    ymin = rf[l].ymin; ymax = rf[l].ymax;
    xmin = rf[l].xmin; xmax = rf[l].xmax;

    for(k=0; k<nbounds; k++){ Bmin[k] = bignum; Bmax[k] = -bignum; }

    // First scan, compute bounds
    for (z = zmin; z <= zmax; z++)
      for (y = ymin; y <= ymax; y++)
	for (x = xmin; x <= xmax; x++)
	  if (label.at(x,y,z)==l && isEdge(label, x, y, z) )
	    for(i = 0; i < nbounds; i++) {
	      v = nx[i]*x + ny[i]*y + nz[i]*z;
	      Bmin[i] = min(v, Bmin[i]);
	      Bmax[i] = max(v, Bmax[i]);
	    }
    
    // Second scan
    for (z = zmin; z <= zmax; z++)
      for (y = ymin; y <= ymax; y++)
	for (x = xmin; x <= xmax; x++)
	  if (label.at(x,y,z) < 0 ) {
	    for(i = 0, added = true; i < nbounds; i++)
	      if ((v = nx[i]*x + ny[i]*y + nz[i]*z) < Bmin[i] || v > Bmax[i]) {
		added = false;
		break;
	      }
	    if (added) label.set(x,y,z, l);
	  } 
  }
 
}

ShortImage __convex_hull_3d__(const ByteImage &img, const int N)
{
  if (img.isNull()) return ShortImage();

  int nlabel;
  ShortImage label = bip::regions(img, nlabel);
   __convex_hull_3d__(label, nlabel, N);
  return label;
}


ShortImage bip::convexHull(const ByteImage &img, int N)
{
  N = max(3, min(50, N));
  if (img.isNull()) return ShortImage();
  if (img.nz()>1) return __convex_hull_3d__(img, N);
  return __convex_hull_2d__(img, N);
}

void bip::convexHull(ShortImage &img, int N)
{
  N = max(3, min(50, N));
  if (img.isNull()) return;
  
  int nlabel = bip::regionNumber(img);
  if (img.nz()>1)
    __convex_hull_3d__(img, nlabel, N);
  else    
    __convex_hull_2d__(img, nlabel, N);
}


void bip::selectRegionsBySize(ByteImage &img, int min_size, int max_size)
{
  if (img.isNull() || min_size<=0 ) return;
  if (max_size < 0 ) max_size = img.npix()+1;
  int nlabel;
  ShortImage label = bip::regions(img, nlabel);
  vector<RegionFeature> rf = bip::regionFeatures(label, nlabel);
  if (rf.empty() || nlabel != rf.size()) return;

  vector<bool> remove(nlabel);  
  for(int i = 0; i < nlabel; i++) 
    remove[i] = (rf[i].npix < min_size || rf[i].npix > max_size  );
  
  const int npix = img.npix();

  for(int i = 0; i < npix; i++)
    if (label.at(i))
      if (remove[label.at(i)])  img.set(i, byte(Background));
}
 

void bip::selectRegions(ByteImage &img, const Image &br, float min_int, int min_sz)
{
  if (img.isNull()) return;
  int nlabel;
  ShortImage label = bip::regions(img, nlabel);
  vector<RegionFeature> rf = bip::regionFeatures(label, nlabel, br);
  if (rf.empty()){
    cerr << "bip::regionFeatures failed." << endl;
    return;
  } 
  vector<bool> remove(nlabel);
  for(int i = 0; i < nlabel; i++)
    remove[i] = (rf[i].npix < min_sz || rf[i].mean < min_int);
 
  const int npix = img.npix();
  for(int i = 0; i < npix; i++)
    if (label.at(i)>=0)
      if (remove[label.at(i)]) img.set(i, Background);
}


FloatImage bip::fchamfer(const Image &img)  
{
  if (img.isNull()) {
    cerr << "In bip::chamfer" << endl;
    cerr << "   Warning: input is empty, return null image" <<  endl; 
    return FloatImage();
  }
  
  FloatImage D;
  if (img.format() == BYTE)
    D =  __init__chamfer_byte__(*((ByteImage *) &img) );
  else if (img.format() == SHORT)
    D =  __init__chamfer_short__(*((ShortImage *) &img) );
  else {
    cerr << "In bip::chamfer" << endl;
    cerr << "   Warning: input is not BYTE or SHORT image, return null image"<< endl;
    return FloatImage();    
  }

  if (D.nz() > 1)  __compute_chamfer_3d__ (D);
  else __compute_chamfer_2d__ (D);
  return D;
}


ByteImage bip::chamfer(const Image &img, bool for_watershed)  
{
  if (img.isNull()) {
    cerr << "In bip::chamfer" << endl;
    cerr << "   Warning: input is empty, return null image" <<  endl; 
    return ByteImage();
  }
  
  FloatImage D;
  if (img.format() == BYTE)
    D =  __init__chamfer_byte__(*((ByteImage *) &img) );
  else if (img.format() == SHORT)
    D =  __init__chamfer_short__(*((ShortImage *) &img) );
  else {
    cerr << "In bip::chamfer" << endl;
    cerr << "   Warning: input is not BYTE or SHORT image, return null image"<< endl;
    return ByteImage();    
  }

  if (D.nz() > 1)  __compute_chamfer_3d__ (D);
  else __compute_chamfer_2d__ (D);
  ByteImage B = image_io::convertToByteImage(D, true);
  if (!for_watershed)  return B;
  const int npix(B.npix());
  for(int k = 0; k < npix; k++)  B.set(k, 255-B.at(k));
  return B;
}

IntArray bip::histogram(const ByteImage & img, int min_int, int max_int)
{
  IntArray h(256, 0);
  const int npix(img.npix());
  int k;
  for(int i=0; i<npix; i++) 
    if ((k=int(img.at(i)))>=min_int && k<=max_int) h[k]++;
  return h;
}


vector<float> RegionFeature::meanArray(const vector<RegionFeature> & fa)
{
  vector<float> a(fa.size());
  for(int i=0; i<a.size(); i++) a[i] = fa[i].mean;
  return a;
}

vector<float> RegionFeature::varArray(const vector<RegionFeature> & fa)
{
  vector<float> a(fa.size());
  for(int i=0; i<a.size(); i++) a[i] = fa[i].var;
  return a;
}

IntArray RegionFeature::npixArray(const vector<RegionFeature> & fa)
{
  IntArray a(fa.size());
  for(int i=0; i<a.size(); i++) a[i] = fa[i].npix;
  return a;
}


void bip::removeInternalStructures(ByteImage &img)
{
  const int nx = img.nx();
  const int ny = img.ny();
  const int nz = img.nz();
  ByteImage I(nx+2, ny+2, nz==1? 1 : nz+2);
  I.reset(byte(Background));
  int x, y, z;
  if (nz==1) 
    for(y=0; y<ny; y++)
      for(x=0; x<nx; x++)
	I.set(x+1, y+1, img.at(x, y));
  else 
    for(z=0; z<nz; z++)
      for(y=0; y<ny; y++)
	for(x=0; x<nx; x++)
	  I.set(x+1, y+1, z+1, img.at(x, y, z));

  __remove__internal_structures__(I);  

  if (nz==1) 
    for(y=0; y<ny; y++)
      for(x=0; x<nx; x++)
	img.set(x, y, I.at(x+1, y+1) );
  else 
    for(z=0; z<nz; z++)
      for(y=0; y<ny; y++)
	for(x=0; x<nx; x++)
	  img.set(x, y, z, I.at(x+1, y+1, z+1) );
}




ShortImage bip::regions(const ShortImage &img)
{
  int nc;
  return bip::regions(img, nc);
}

ShortImage bip::regions(const ShortImage &img, int &nlabel)
{  
  if (img.isNull()) return ShortImage();
 
  const int nx = img.nx();
  const int ny = img.ny();
  const int nz = img.nz();
  const int npix = img.npix();
  const int nx_1 = nx-1;
  const int ny_1 = ny-1;
  const int nz_1 = nz-1;
  
  ShortImage label(nx, ny, nz); 
  label.reset(-2);
 
  
  int x, y, z,  k, x_1, x0, x1, y_1, y0, y1, z_1, z0, z1, xx, yy, zz; 
  int cur_int;
  queue <Point3D> Q;
  Point3D q;
  
  nlabel = 0;
  for(z = 0; z < nz; z++)     
    for(y = 0; y < ny; y++) 
      for(x = 0; x < nx; x++) 
	if (label.at(x, y, z) == -2) {
	  
	  Q = queue<Point3D>();
	  Q.push(Point3D(x,y,z));
	  cur_int = img.at(x, y, z); 
	  while (!Q.empty()) {
	    
	    q = Q.front();
	    xx = q.x(); yy = q.y(); zz = q.z();
	    label.set(xx, yy, zz, nlabel);
	    
	    if ( (x_1 = xx - 1) < 0 ) x_1 = 0;
	    if ( (x1 = xx + 1) >= nx ) x1 = nx_1;
	    if ( (y_1 = yy - 1) < 0 ) y_1 = 0;
	    if ( (y1 = yy + 1) >= ny ) y1 = ny_1;
	    if ( (z_1 = zz - 1) < 0 ) z_1 = 0;
	    if ( (z1 = zz + 1) >= nz ) z1 = nz_1;	    
	    
	    for(z0 = z_1; z0 <= z1; z0++)
	      for(y0 = y_1; y0 <= y1; y0++)
		for(x0 = x_1; x0 <= x1; x0++)
		  if (label.at(x0, y0, z0) == -2 && img.at(x0, y0, z0)==cur_int) {   
		    Q.push(Point3D(x0, y0, z0));
		    label.set(x0, y0, z0, nlabel);
		  }	      	    
	    
	    Q.pop();
	    
	  }  
	  
	  nlabel++;
	} 

  return label;
}


vector<PointArray> bip::boundaries(const ShortImage &img)
{
  ByteImage E = bip::edge(img);
  const int nr = regionNumber(img);
  vector<PointArray> B(nr);
  IntArray c(nr, 0);
  const int npix = img.npix();
  const int nx = img.nx();
  const int ny = img.ny();

  for(int i=0; i<npix; i++)
    if (img.at(i)>=0&&E.at(i)==Foreground)
      c[img.at(i)]++;
  
  for(int i=0; i<nr; i++) {
    B[i] = PointArray();
    B[i].reserve(c[i]);
  }
    
  for(int y=0; y<ny; y++)
    for(int x=0; x<nx; x++)
      if (img.at(x, y)>=0&&E.at(x, y)==Foreground)
	B[img.at(x, y)].push_back(Point(x, y));

  return B;
}



void  bip::relabel(ShortImage &I) 
{  
  if (I.isNull()) return; 
  int i;
  const int num = I.npix();
  short *p = I.ptr(); 
  const int n = regionNumber(I);

  IntArray m(n, 0);  
  for(i = 0; i < num; i++)
    if (p[i] >= 0 && p[i] < n) m[p[i]]++;
  
  IntArrayIterator it;
  for(it = m.begin(), i = 0; it != m.end(); it++) 
    if (*it > 0) {
      *it = i; i++;
    } else *it = -1;
  
  for(i = 0; i < num; i++)
    if (p[i] >= 0 && p[i] < n)
      p[i] = m[p[i]];
}

void bip::reverse(ByteImage &I)
{
  if (I.isNull()) return; 
  const int npix = I.npix();
  for(int i=0; i<npix; i++) 
    I(i) = 255-I(i);
}

vector<FloatArray> bip::regionIntensityArrays(const ShortImage &label, 
					      const Image &br, bool sorted)
{
  return bip::regionIntensityArrays(label, bip::regionNumber(label), br, sorted);
}

vector<FloatArray> bip::regionIntensityArrays(const ShortImage &label, 
					      int nc, 
					      const Image &br, bool sorted)
{
  if (label.isNull() || !label.equalSize(br))
    return vector<FloatArray>();

  IntArray npixs(nc, 0);
  const int npix = label.npix();
  int k;
  for(int i=0; i<npix; i++)
    if ( (k=label.at(i))>=0 && k<nc) npixs[k]++;
    
  vector<FloatArray> a(nc);
  for(int i=0; i<nc; i++) a[i].reserve(npixs[i]);

  for(int i=0; i<npix; i++)
    if ( (k=label.at(i))>=0 && k<nc) a[k].push_back(br.pixel(i));

  if (sorted) 
    for(int i=0; i<nc; i++) sort(a[i].begin(), a[i].end());
  return a;
}

namespace {

  void __propagate_3d__(ShortImage &img, const float max_dist)  
  {
    if (img.isNull()) return;  
    const float max_val = 10000;
    const int nx = img.nx();  
    const int ny = img.ny();
    const int nz = img.nz();
    const int nx_1 = nx-1;
    const int ny_1 = ny-1;
    const int nz_1 = nz-1;
    const int nx_2 = nx-2;
    const int ny_2 = ny-2;
    const int nz_2 = nz-2;
  
    const int npix = img.npix();
    int  i, j, z, k;
    float min_val;
    int kmin;
  
    // Initialization
    FloatImage D(nx, ny, nz);
    for (i = 0; i < npix; i++) 
      if (img.at(i)>=0)  D.set(i, 0);
      else D.set(i, max_val);
    
    IntArray _off;
    int _off_size;
    vector<float> _off_dist;
    IntArrayIterator it;
    float d;
  
    //  cout << "1st scan..." ; 
    _off = IntArray();
    _off_dist = vector<float>();
    _off_dist.reserve(27);
    _off.reserve(27);
    for(z = -1; z <= 1; z++)
      for(i = -1; i <= 1; i++)
	for(j = -1; j <= 1; j++) 
	  if ((k = z*ny*nx + i*nx+j) != 0 && i+j+z<=0 ) {
	    _off.push_back(k);
	    d = sqrt(float(z*z+i*i+j*j));
	    _off_dist.push_back(d);
	  }

    int loc;
    _off_size = _off.size();
    for(z = 1; z < nz_1; z++) 
      for (i = 1; i < ny_1; i++)
	for (j = 1; j < nx_1; j++) 
	  if ( D.at(loc=D.pos(j, i, z)) > 0 ){	  
	    min_val = max_val + 2;
	    kmin = -1;
	    for(k = 0; k < _off_size; k++)
	      if ( (d=D.at(loc+_off[k])+_off_dist[k]) < min_val) {
		min_val = d;
		kmin = k;
	      }
	  
	    if (min_val >= max_val) continue; 
	    if (min_val  < D.at(loc)) {
	      if (kmin >= 0) img.set(loc, img.at(loc+_off[kmin]));	    
	      D.set(loc, min_val);
	    }
	  }
  
    //  cout << "2nd scan..." ;
    _off = IntArray();
    _off.reserve(27);
    _off_dist = vector<float>();
    _off_dist.reserve(27);
    for(z = -1; z <= 1; z++)
      for(i = -1; i <= 1; i++)
	for(j = -1; j <= 1; j++) 
	  if ((k = z*ny*nx + i*nx+j) != 0 && i+j+z>=0 ) {
	    _off.push_back(k);
	    d = sqrt( float(z*z+i*i+j*j) );
	    _off_dist.push_back(d);
	  }
  
    _off_size = _off.size();
    for(z = nz_2; z >= 1; z--)
      for (i = ny_2; i >= 1; i--)
	for (j = nx_2; j >= 1; j--) 
	  if (  D.at(loc=D.pos(j, i, z)) > 0 ){	  
	    min_val = max_val + 2;
	    kmin = -1;
	    for(k = 0; k < _off.size(); k++)
	      if ( (d=D.at(loc+_off[k])+_off_dist[k]) < min_val) {
		min_val = d;
		kmin = k;
	      }
	  
	    if (min_val >= max_val) continue; 
	    if (min_val  < D.at(loc)) {
	      if (kmin >= 0) img.set(loc, img.at(loc+_off[kmin]));	    
	      D.set(loc, min_val);
	    }
	  }

    for(i=0; i<npix; i++) if (D.at(i)>=max_dist) img.set(i, -1);
  }

  void __propagate_2d__(ShortImage &img, const float max_dist)  
  {  
    if (img.isNull()) return;  
    const float max_val = 10000;
    const int nx = img.nx();  
    const int ny = img.ny();
    const int nx_1 = nx-1;
    const int ny_1 = ny-1;
    const int nx_2 = nx-2;
    const int ny_2 = ny-2;
    const int npix = nx*ny;

    int  i, j, k;
    float min_val;
    int kmin;
    
    // Initialization
    FloatImage D(nx, ny);
    for (i = 0; i < npix; i++) 
      if (img(i)>=0) D(i) = 0;
      else D(i) = max_val;
    
    IntArray _off;
    int _off_size;
    vector<float> _off_dist;
    IntArrayIterator it;
    float d;
  
    //  cout << "1st scan..." ; 
    _off = IntArray();
    _off_dist = vector<float>();
    _off_dist.reserve(9);
    _off.reserve(9);
    for(i = -1; i <= 1; i++)
      for(j = -1; j <= 1; j++) 
	if ((k = i*nx+j) != 0 && i+j<=0 ) {
	  _off.push_back(k);
	  d = sqrt(double(i*i+j*j));
	  _off_dist.push_back(d);
	}
    
    int loc;
    _off_size = _off.size();
    
    for (i = 1; i < ny_1; i++)
      for (j = 1; j < nx_1; j++) 
	if ( D(loc=D.pos(j, i)) > 0.5 ){	  
	  min_val = max_val + 2;
	  kmin = -1;
	  for(k = 0; k < _off_size; k++)
	    if ( (d=D(loc+_off[k])+_off_dist[k]) < min_val) {
	      min_val = d;
	      kmin = k;
	    }
	  
	  if (min_val >= max_val) continue; 
	  if (min_val  < D(loc)) {
	    D(loc) = min_val;
	    if (kmin >= 0) 
	      img(loc) = img(loc+_off[kmin]);

	  }
	}
  
    //  cout << "2nd scan..." ;
    _off = IntArray();
    _off.reserve(9);
    _off_dist = vector<float>();
    _off_dist.reserve(9);
    for(i = -1; i <= 1; i++)
      for(j = -1; j <= 1; j++) 
	if ((k = i*nx+j) != 0 && i+j>=0 ) {
	  _off.push_back(k);
	  d = sqrt( float(i*i+j*j) );
	  _off_dist.push_back(d);
	}
  
    _off_size = _off.size();
    for (i = ny_2; i >= 1; i--)
      for (j = nx_2; j >= 1; j--) 
	if (  D(loc=D.pos(j, i)) > 0 ){	  
	  min_val = max_val + 2;
	  kmin = -1;
	  for(k = 0; k < _off_size; k++)
	    if ( (d=D(loc+_off[k])+_off_dist[k]) < min_val) {
	      min_val = d;
	      kmin = k;
	    }
	  
	  if (min_val >= max_val) continue; 
	  if (min_val  < D(loc)) {
	    if (kmin >= 0) 
	      img(loc) = img(loc+_off[kmin]); 
	    D(loc) = min_val;
	  }
	}

    for(i=0; i<npix; i++) if (D.at(i)>=max_dist) img(i) = -1;
  }
  
}


void bip::propagate(ShortImage &img, const float max_dist)  
{
  if (img.isNull()) return;
  if (img.nz()>1) {
    __propagate_3d__(img, max_dist);
    return;
  }
  
  __propagate_2d__(img, max_dist);
  return;
}


void bip::erode(ByteImage &img) 
{
  if (img.isNull()) return;
  const int nx = img.nx();
  const int ny = img.ny();
  const int npix = nx * ny;

  int d, i, j, x, y;

  IntArray off;
  off.reserve(8);  
  for(y=-1; y<=1; y++)
    for(x=-1; x<=1; x++) {
      d = y*nx+x;
      if (d!=0) off.push_back(d);
    }
  const int noff = off.size();
  vector<bool> interior = img.interior();
  deque<int> p;

  for(i=0; i<npix; i++) 
    if (img(i)==Foreground) {
      if (!interior[i]) p.push_back(i);
      else  
	for(j=0; j<noff; j++)
	  if (img(i+off[j])!=Foreground) {
	    p.push_back(i);
	    break;
	  }
    }
  
  j = p.size();
  for(i=0; i<j; i++) 
    img(p[i]) = Background;
}


void bip::dilate(ByteImage &img) 
{
  if (img.isNull()) return;
  const int nx = img.nx();
  const int ny = img.ny();
  const int npix = nx * ny;
  int d, i, j, x, y;
  IntArray off;
  off.reserve(8);  
  for(y=-1; y<=1; y++)
    for(x=-1; x<=1; x++) {
      d = y*nx+x;
      if (d!=0) off.push_back(d);
    }
  const int noff = off.size();
  vector<bool> interior = img.interior();

  ByteImage _img = img.duplicate();
  for(i=0; i<npix; i++) 
    if (_img(i)==Foreground && interior[i])
      for(j=0; j<noff; j++)
	if (_img(d=i+off[j])==Background ) 
	  img(d) = Foreground;
}


void bip::erode(ShortImage &img) 
{
  if (img.isNull()) return;
  const int nx = img.nx();
  const int ny = img.ny();
  const int npix = nx * ny;
  
  int d, i, j, x, y;

  IntArray off;
  off.reserve(8);  
  for(y=-1; y<=1; y++)
    for(x=-1; x<=1; x++) {
      d = y*nx+x;
      if (d!=0) off.push_back(d);
    }
  const int noff = off.size();
  vector<bool> interior = img.interior();
  deque<int> p;

  int br;

  for(i=0; i<npix; i++) 
    if ((br=img(i))>=0) {
      if (!interior[i]) p.push_back(i);
      else  
	for(j=0; j<noff; j++)
	  if (img(i+off[j])!=br) {
	    p.push_back(i);
	    break;
	  }
    }
  
  j = p.size();
  for(i=0; i<j; i++) img(p[i]) = -1;
}


