#include <jni.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "./circularvoting.h"
#include "./scalespace.h"

using namespace std;
using namespace stdi;
using namespace voting;

#include "JNICircularVoting.h"

extern JNIEXPORT jintArray JNICALL Java_JNICircularVoting_compute (JNIEnv *env, 
    jobject jobj, 
    jintArray jPixels, 
    jint jRow, 
    jint jCol, 
    jint jHMin,
    jint jHMax,
    jint jRadius, 
    jint jMinGrad,
    jint jMinVote, 
    jint jSigma)
{
    jintArray jb;
    jint *buf, *pixels;

    ByteImage* img = new ByteImage(jCol, jRow);
    int i, num;
    num = jCol * jRow;
    pixels = new jint[num];
    env->GetIntArrayRegion(jPixels, 0, num, pixels);
    for (i = 0; i < num; i++){
        img->set(i, (byte)(pixels[i]));
    }

    FloatImage grad = scalespace::gradient(*img, jSigma, 1);
    float min_grad = grad.fmax() * 0.1;
    int hmin = 1;
    int hmax = jHMax;
    int radius = (int)(hmax * 0.25 + 0.5);
    float min_vote = (float)jMinVote;
    float sigma = (float)jSigma;

    CircularVoting cv;
    cv.setParams(CircularVoting::BrightReg, hmin, hmax, radius, min_grad, 
                 min_vote, sigma, CircularVoting::PowMethod, 
                 (float) 1, false, true);
    
    cv.saveHistory(false);
    cv.setPrefix("tmp/");
    cv.compute(*(img->frame(0)));
    PointArray seeds = cv.centers();
    vector<float> weights = cv.center_weights();
    //cout << "Found " << seeds.size() << " seeds." << endl;
    
    if (seeds.size() < 1){
        num = 1; 
        buf = new jint[num * 3];
        buf[0] = jRow / 3;
        buf[1] = jCol / 3;
        buf[2] = 0;
    }else{
        num = seeds.size();
        buf = new jint[num * 3];
        for (i = 0; i < num; i++){
            buf[i * 3] = seeds[i].y();
            buf[i * 3 + 1] = seeds[i].x();
            buf[i * 3 + 2] = (int)weights[i];
        }
    }
    
    jb = env->NewIntArray(3 * num);
    env->SetIntArrayRegion(jb, 0, 3 * num, buf);
    
    delete[] buf;
    delete[] pixels;

    return (jb);
}

