import ij.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import javax.swing.JFrame;

public class CircularVoting_ implements PlugInFilter {
    int channel;
    int hmin;
    int hmax;
    int radius;
    int minGrad;
    int minVote;
    int sigma;
    
    public int setup(String arg, ImagePlus imp) {
        if (arg.equals("about")){
            showAbout(); 
            return DONE;
        }
        
        return DOES_RGB + DOES_8G;
    }

    public void run(ImageProcessor ip) {
        boolean isByte = ip instanceof ByteProcessor;
	if (showArgWindow(isByte)){
	    // get width, height and the region of interest
            int w = ip.getWidth();     
	    int h = ip.getHeight();    
	   
            Rectangle roi = ip.getRoi();

            int[] cpixels;
            if (!isByte){
		int[] pixels = (int[]) ip.getPixels();
		int shift = channel * 8;
                cpixels = new int[pixels.length];
		for (int i = 0; i < pixels.length; i++){
		    int p = (pixels[i] >> shift) & 0xFF;
		    cpixels[i] = (p << 16) + (p << 8) + p;
		}
	    }else{
		byte[] pixels = (byte[]) ip.getPixels();
                cpixels = new int[pixels.length];
		for (int i = 0; i < pixels.length; i++){
		    int p = pixels[i];
		    cpixels[i] = (p << 16) + (p << 8) + p;
		}
	    }
            int size = 4;
            JNICircularVoting ct = new JNICircularVoting();
            int[] cor = ct.compute(cpixels, h, w, hmin, hmax, radius, minGrad, minVote, sigma);
            if (cor.length < 2){
                IJ.error("NO seeds Found.");
	    }else{
                showResultUI(cor);

		if (isByte){
		    byte[] pixels = (byte[]) ip.getPixels();
                    for (int k = 0; k < cor.length / 3; k++){

                        //found color first;
                        int c = 0;
			for (int i= cor[k * 3] - size; i < cor[k * 3] + size; i++){
                            int pos = i * w + cor[k * 3 + 1];
			    if (pixels[pos] > 0){
                                c++;
                            }
                        }
			for (int i= cor[k * 3 + 1] - size; i < cor[k * 3 + 1] + size; i++){
                            int pos = cor[k * 3] * w + i;
			    if (pixels[pos] > 0){
                                c++;
                            }
                        }
       
                        byte color = (byte)0x00;
                        if (c > size * 2){
                            color = (byte)0xFF;
			}
                       
			for (int i= cor[k * 3] - size; i < cor[k * 3] + size; i++){
                            int pos = i * w + cor[k * 3 + 1];
			    pixels[pos] = color;
                        }
			for (int i= cor[k * 3 + 1] - size; i < cor[k * 3 + 1] + size; i++){
                            int pos = cor[k * 3] * w + i;
			    pixels[pos] = color;
			}
		    }
                }else{
	  	    int[] pixels = (int[]) ip.getPixels();
                    for (int k = 0; k < cor.length / 3; k++){
			for (int i= cor[k * 3] - size; i < cor[k * 3] + size; i++){
                            int pos = i * w + cor[k * 3 + 1];
			    pixels[pos] = 0xFF0000;
                        }
			for (int i= cor[k * 3 + 1] - size; i < cor[k * 3 + 1] + size; i++){
                            int pos = cor[k * 3] * w + i;
			    pixels[pos] = 0xFF0000;
                        }
		    } 
		}
	    }
	}
    }

    void showAbout() {
        IJ.showMessage("Circular Voting", "Lawrence Berkeley National Lab");
    }

    boolean showArgWindow(boolean isByte){
        GenericDialog gd = new GenericDialog("Circular Voting settings");

        if (!isByte){
            gd.addNumericField("Channel:" , 0, 0);
        }
        gd.addNumericField("Diameter:", 26, 0);
        gd.addNumericField("Min Vote:", 1000, 0);
        gd.addNumericField("Sigma:", 3, 0);

        gd.showDialog();
        if (gd.wasCanceled()){
            IJ.error("Plugin canceled!");
            return false;
	}
       
        if (!isByte){
            channel = (int)gd.getNextNumber();
            if (channel < 0){
                channel = 0;
	    }
            if (channel > 2){
                channel = 2;
	    }
	}

        hmin = 0;
        hmax = (int)(gd.getNextNumber() * 0.75);
        radius = 0;
        minGrad = 0;
        minVote = (int)gd.getNextNumber();
        sigma = (int)gd.getNextNumber();

        return true;
    }

    void showResultUI(int[] results){
        JFrame frame = new JFrame("CircularVoting");
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        int num = results.length / 3;
        Object[][] data = new Object[num][4];
        for (int i = 0; i < num; i++){
            data[i][0] = new Integer(i);
            data[i][1] = new Integer(results[i * 3 + 1]);
            data[i][2] = new Integer(results[i * 3]);
	    data[i][3] = new Integer(results[i * 3 + 2]);
	}
        CVTable newContentPane = new CVTable(data);
        newContentPane.setOpaque(true); //content panes must be opaque
        frame.setContentPane(newContentPane);

        frame.pack();
        frame.setVisible(true);
    }

}

